import matplotlib.pyplot as plt
import librosa.display
import librosa as lr
import os

__all__ = ['quickPlotSpectrogram']


def quickPlotSpectrogram(signal, sample_rate=22050):
    """
    Draw a spectrogram of the specified signal

    Parameters
    ------
    `y`:
        Array, String. Path to audio or audio data.
    `sample_rate`:
        Desired sample rate of audio. Default is `22050`.
    """

    if os.path.isfile(signal):
        x, _ = lr.load(signal, sr=sample_rate)
    else:
        x = signal    

    X = lr.stft(x)
    Xdb = lr.amplitude_to_db(abs(X))
    plt.figure(figsize=(14, 5))
    librosa.display.specshow(Xdb, sr=sample_rate, x_axis='time', y_axis='hz')
    plt.colorbar()
    plt.show()


from augmolino import augmenter, augmentation
import os

source_dir = "tests/sounds/"
source_file = "tests/sounds/speech.wav"


def test_empty_augmenter_init():

    a = augmenter.augmenter()
    assert a.pipe == []


def test_filled_augmenter_init():

    augs = [
        augmentation.pitchShift(semitones=1),
        augmentation.pitchShift(semitones=2)]

    a = augmenter.augmenter(augs)

    assert len(a.pipe) == len(augs)

    for aug in a.pipe:
        assert aug.descriptor == "pitch_shift"


def test_augmenter_summary():

    augs = [
        augmentation.pitchShift(semitones=1),
        augmentation.pitchShift(semitones=2)]

    a = augmenter.augmenter(augs)

    a.summary()


def test_augmenter_run_array():

    augs = [
        augmentation.timeStretch(rate=2),
        augmentation.pitchShift(semitones=2),
        augmentation.offsetAudio(s=1)]

    a = augmenter.augmenter(augs)

    xs = a.execute(source_file)

    assert len(xs) == len(augs)


def test_augmenter_run_single_save():

    fnames = ["targetfile1.wav",
              "targetfile2.wav",
              "targetfile3.wav"]

    augs = [
        augmentation.timeStretch(rate=2),
        augmentation.pitchShift(semitones=2),
        augmentation.offsetAudio(s=1)]

    a = augmenter.augmenter(augs)

    for fname in fnames:
        xs = a.execute(source_file, fname)
        assert xs == None

    for fname in fnames:
        assert os.path.exists(fname)
        os.remove(fname)


def test_augmenter_run_multi_save():

    fnames = ["targetfile1.wav",
              "targetfile2.wav",
              "targetfile3.wav"]

    augs = [augmentation.timeStretch(rate=2),
            augmentation.pitchShift(semitones=2),
            augmentation.offsetAudio(s=1)]

    a = augmenter.augmenter(augs)

    for fname in fnames:
        xs = a.execute(source_file, fname)
        assert xs == None

    for fname in fnames:
        assert os.path.exists(fname)
        os.remove(fname)

def test_augmenter_run_folder_auto_save():

    augs = [augmentation.timeStretch(rate=2),
            augmentation.pitchShift(semitones=2),
            augmentation.offsetAudio(s=1)]

    a = augmenter.augmenter(augs)

    xs = a.execute(source_dir, dest="auto")
    assert xs == None

def test_cleanup():
    keep = ["impulse_response.wav",
            "speech.wav"]

    for file in os.listdir(source_dir):
        if file not in keep:
            os.remove(os.path.join(source_dir, file))

# this can only be tested visually by a human
test_augmenter_summary()



import os
import librosa as lr
import augmolino.augmentation as aug
import numpy as np
import numpy.testing as npt

test_file = "tests/sounds/impulse_response.wav"
target_path = "tests/sounds/mysound.wav"


def test_augmentation_init():

    a = aug._augmentation()

    assert a.sample_rate == 22050
    assert a.descriptor == aug.descriptors["_augmentation"]
    assert a.function == None
    assert len(a.kwargs) == 0
    assert a.tag == ""


def test_augmentation_load():

    a = aug._augmentation()
    a._load(test_file)

    assert a.f_source == test_file

    t, _ = lr.load(path=test_file)

    assert len(a.signal) == len(t)


def test_augmentation_autoName():
    a = aug._augmentation()
    a._load(test_file)

    assert a.tag == ""

    name = a._autoName(a.descriptor, a.tag)

    assert name == test_file[:-4] + f"_{a.descriptor}_{a.tag}.wav"


def test_timeStretch_class():

    test_rate = 0.5
    a = aug.timeStretch(rate=test_rate)

    assert a.descriptor == aug.descriptors[aug.__all__[0]]
    assert a.function == lr.effects.time_stretch
    assert len(a.kwargs) == 1
    assert a.tag == "0_5"

    x = a.run(test_file)

    assert a.f_dest == None
    assert len(x) == int(len(a.signal) / test_rate)
    npt.assert_allclose(x, lr.effects.time_stretch(a.signal, rate=test_rate))

    x = a.run(test_file, target_path)

    assert x == None
    assert a.f_dest == target_path
    assert os.path.exists(target_path)

    os.remove(target_path)

    x = a.run(test_file, f_dest="auto")
    assert a.f_dest == test_file[:-4] + f"_{a.descriptor}_{a.tag}.wav"
    assert os.path.exists(a.f_dest)

    os.remove(a.f_dest)
    

def test_pitchShift_class():

    test_semitones = 2
    a = aug.pitchShift(semitones=test_semitones)

    assert a.descriptor == aug.descriptors[aug.__all__[1]]
    assert a.function == lr.effects.pitch_shift
    assert len(a.kwargs) == 2
    assert a.tag == "22050_2"

    x = a.run(test_file)

    assert len(x) == len(a.signal)
    npt.assert_allclose(x, lr.effects.pitch_shift(a.signal, 
                                                   sr=a.sample_rate,
                                                   n_steps=test_semitones))

    x = a.run(test_file, f_dest="auto")
    assert a.f_dest == test_file[:-4] + f"_{a.descriptor}_{a.tag}.wav"
    assert os.path.exists(a.f_dest)

    os.remove(a.f_dest)


def test_offsetAudio_class():

    test_s = 1
    a = aug.offsetAudio(s=test_s)

    assert a.descriptor == aug.descriptors[aug.__all__[2]]
    assert a.function == a._offset
    assert len(a.kwargs) == 1
    assert a.tag == "1"

    x = a.run(test_file)

    assert len(x) == len(a.signal) + test_s * a.sample_rate
    npt.assert_allclose(x[0:test_s * a.sample_rate],
                        np.zeros(test_s * a.sample_rate))

    test_s = -1
    a = aug.offsetAudio(s=test_s)

    x = a.run(test_file)
    assert len(x) == len(a.signal) + test_s * a.sample_rate

    x = a.run(test_file, f_dest="auto")
    assert a.f_dest == test_file[:-4] + f"_{a.descriptor}_{a.tag}.wav"
    assert os.path.exists(a.f_dest)

    os.remove(a.f_dest)


def test_fadeAudio_class():

    test_s = 1
    test_direction = "in"
    a = aug.fadeAudio(s=test_s, direction=test_direction)

    assert a.descriptor == aug.descriptors[aug.__all__[3]]
    assert a.function == a._fade
    assert len(a.kwargs) == 2
    assert a.tag == "1_in"

    x = a.run(test_file)


def test_mixAudio_class():
    test_f_mix = "tests/sounds/speech.wav"
    test_ratio = 0.5
    test_start_at = 1
    a = aug.mixAudio(test_ratio, test_start_at)

    assert a.descriptor == aug.descriptors[aug.__all__[4]]
    assert a.function == a._mix
    assert len(a.kwargs) == 2
    assert a.tag == "0_5_1"

    x = a.run(test_file, target_path, f_mix=test_f_mix)
    assert len(a.kwargs) == 3

    a = aug.mixAudio(test_ratio)
    assert a.tag == "0_5_None"
    x = a.run(test_file, target_path, f_mix=test_f_mix)


    os.remove(target_path)


from augmolino import augmentation
import numpy as np
import os


class augmenter:

    def __init__(self, augmentations=None):
        """
        Group class which holds a dynamic amount of 
        augmentations specified by the user

        Parameters
        ----------
        `augmentations`:
            Augmentation. Single or array of augmentations from
            `augmentation.py`. If left unspecified, the augmenter
            can later be filled with augmentations via `augmenter.add()`.
            Default is `None`.

        Examples
        --------
        >>> # get augmentations as signals and plot them:
        >>> from augmolino import augmenter, augmentation
        >>> import matplotlib.pyplot as plt
        >>> # specify set of augmentations:
        >>> augs = [
                augmentation.timeStretch(rate=2),
                augmentation.pitchShift(semitones=2),
                augmentation.offsetAudio(s=1)]
        >>> # create the augmenter
        >>> a = augmenter.augmenter(augs)
        >>> # run augmenter
        >>> xs = a.execute("tests/sounds/impulse_response.wav")
        >>> # create plot
        >>> fig, axs = plt.subplots(3,1)
        >>> # display signals
        >>> for i, x in enumerate(xs):
            >>> axs[i].plot(x)
        >>> plt.show()
        """

        if not augmentations:
            # init empty augmenter
            self.pipe = []

        else:
            # create array of augmentations
            if len(augmentations) > 1:
                self.pipe = augmentations
            else:
                self.pipe = [augmentations]

    def add(self, augmentation):
        """
        Add a single augmentation to the augmenter pipeline.

        Parameters
        ----------
        `augmentation`:
            Augmentation. Gets appended to existing or empty
            pipe of augmentations within the augmenter.
        """
        self.pipe.append(augmentation)

    def execute(self, source, dest=None, **kwargs):
        """
        Run all augmentations within the pipe. Specific settings are
        inside of each augmentation.

        Parameters
        ----------
        `source`:
            List, String. Source of data. Can be a folder,
            array of filenames or single files.
        `dest`:
            String. Destination for saved augmented files. If `None`
            is passed, the first augmented signals per augmentation
            are returned as `numpy` array and not saved. Pass `"auto"`
            to save them to the source folder under unique IDs generated
            from the augmentations settings. Default is `None`.
        `kwargs`:
            Keyword arguments which were not passed in the initialization
            of augmentations. They get appended here.

        Returns
        -------
        `xs`:
            Array, None. Returns each augmented signal if no save location
            has been specified for the coresponding augmentation.
            Otherwise it returns `None`.

        """

        # source given as a list of file paths
        if isinstance(source, (list, tuple, np.ndarray)):
            files = source

        # single source path to file
        if os.path.isfile(source):
            files = [source]

        # source given as parent folder
        if os.path.isdir(source):
            if dest == None:
                raise ValueError("Augmenter without save directory "\
                    "pointed to multiple files. Did you forget "\
                    "to specify a path?")
            files = [os.path.join(source, file) for file in os.listdir(source) if file.lower().endswith("wav")]

        if dest == None:
            xs = [[]] * len(self.pipe)
        else:
            xs = None
        # this is sloooow but the only way to append dynamic sizes
        for i, augmentation in enumerate(self.pipe):
            for file in files:
                x = augmentation.run(file, dest, **kwargs)

                if dest == None:
                    xs[i].append(x)
                    xs[i] = np.asarray(xs[i][i])

        return xs

    def summary(self):

        num_aug = len(self.pipe)

        print("")
        print("------------augmenter.summary------------")
        print("-----------------------------------------")
        print(f" number of augmentations: {num_aug}     ")
        print("")
        print(" type:                                   ")

        for aug in self.pipe:
            print(f" > {aug.descriptor}                 ")

        print("------------augmenter.summary------------")
        print("")



from setuptools import setup

setup(name='augmolino',
      packages=['augmolino'],
      version='0.0.4',
      description='augmentation for audio based datasets for machine learning',
      url='https://github.com/jake-is-ESD-protected/augmolino',
      download_url="https://github.com/jake-is-ESD-protected/augmolino/archive/refs/tags/0.0.2.tar.gz",
      author='Jakob Tschavoll',
      author_email='jt@tschavoll.at',
      license='GPL 3.0',
      classifiers=[],
      keywords=["ML", "augmentation", "audio"],
      zip_safe=False,
      install_requires=[
          'numpy',
          'scipy',
          'librosa',
          'soundfile'])



import warnings
import librosa as lr
import numpy as np
import soundfile as sf
import random as rd
import os

__all__ = ['timeStretch', 'pitchShift',
           'offsetAudio', 'fadeAudio', 'mixAudio']

descriptors = {
    "_augmentation": "base class",
    __all__[0]: "time_stretch",
    __all__[1]: "pitch_shift",
    __all__[2]: "time_offset",
    __all__[3]: "time_fade",
    __all__[4]: "sound_mix"}


class _augmentation:
    """
    Private base class for different augmentations.
    Prepares general parameters which every augmentation
    needs. 

    Parameters
    ----------
    `sample_rate`:
        sample rate of audio file (used for resampling). Default
        is `22050`.
    `function`:
        function which applies the augmentation to the signal
        array. Can be used to reference internal functions or
        functions provided by other modules. Default is `None`.
    `kwargs`:
        Arguments for `function`.

    Notes
    -----
    This is a private base class which is not intened to be
    called on its own. Always call an actual augmentation.
    """

    def __init__(self, sample_rate=22050, function=None, **kwargs):

        self.sample_rate = sample_rate
        self.descriptor = descriptors["_augmentation"]
        self.function = function
        self.kwargs = kwargs
        # use all kwargs values to create a unique ID
        kwargs_vals = []
        for arg in kwargs.values():
            if isinstance(arg, str):
                if "/" in arg or "\\" in arg:
                    arg = self._shortenName(arg)
            kwargs_vals.append(str(arg))
        self.tag = ('_'.join(kwargs_vals)).replace('.', '_')

    def run(self, f_source, f_dest=None, **kwargs):
        """
        Apply the augmentation to a specified file and store 
        the result as array or other file.

        Parameters
        ----------
        `f_source`:
            path to file which should be augmented. Must be in wav format.
        `f_dest`:
            path to augmented file which will be saved. If left unspecified,
            a `numpy` array of the augmented signal is returned instead. Pass
            `"auto"` to store in the source with unique names. Default
            is `None`.
        `kwargs`:
            Arguments for executing of `function`.    
        """
        self._load(f_source)
        self.f_dest = self._parsePath(f_dest)

        # append kwargs in case some are appended after init
        self.kwargs = {**self.kwargs, **kwargs}

        x_new = self._executeFunction()

        if self.f_dest != None:
            sf.write(self.f_dest, x_new, self.sample_rate)
            return None

        else:
            return x_new

    def _load(self, f_source):

        path_details = os.path.splitext(f_source)
        extension = path_details[1]

        if extension != ".wav":
            raise ValueError("File type not supported! Use '.wav' instead.")

        self.f_source = f_source

        x, sr = lr.load(path=self.f_source, sr=self.sample_rate)
        self.signal = x

    def _parsePath(self, f_dest):
        # user wants to save the augmentation
        if f_dest != None:
            if self.f_source == f_dest:
                warnings.warn("Source and save name are the same,\
                    original file will be overwritten!")
                return f_dest
            elif f_dest == "auto":
                return self._autoName(self.descriptor, self.tag)
            else:
                return f_dest

        # user wants a temporary array
        else:
            return None

    def _autoName(self, descriptor, tag):
        if self.f_source == None:
            raise ValueError("No source to autoname from!")
        return self.f_source[:-4] + f"_{descriptor}_{tag}.wav"

    def _executeFunction(self):
        return self.function(self.signal, **self.kwargs)

    def _shortenName(self, f_path):
        return os.path.basename(f_path)[:-4]


class timeStretch(_augmentation):
    """
    Stretch or squeeze a sound while retaining pitch.

    Parameters
    ----------
    `rate`:
        Float. Stretch factor defined by speed of playback. A `rate` < 1
        stretches the sound, a `rate` > 1 squeezes it.
        Default is `1`
    `sample_rate`:
        Int. Sample rate of audio file (used for resampling).
        Default is `22050`
    """

    def __init__(self, rate=1, sample_rate=22050):
        self.rate = rate
        super().__init__(sample_rate=sample_rate,
                         function=lr.effects.time_stretch,
                         rate=rate)
        self.descriptor = descriptors[__all__[0]]


class pitchShift(_augmentation):
    """
    Shift a given input signal's pitch by semitones.

    Parameters
    ----------
    `semitones`:
        Float. Number of western scale semitones to shift up or down.
        Default is `1`
    `sample_rate`:
        Int. Sample rate of audio file (used for resampling).
        Default is `22050`
    """

    def __init__(self, semitones=1, sample_rate=22050):
        self.semitones = semitones
        super().__init__(sample_rate=sample_rate,
                         function=lr.effects.pitch_shift,
                         sr=sample_rate,
                         n_steps=semitones)
        self.descriptor = descriptors[__all__[1]]


class offsetAudio(_augmentation):
    """
    Offset a sound by added dead-time or or by later start.

    Parameters
    ----------
    `s`:
        Float. Offset in seconds. s < 0: skip first samples. 
        s > 0: add dead-time to start. Default is `0`.
    `sample_rate`:
        Int. Sample rate of audio file (used for resampling).
        Default is `22050`
    """

    def __init__(self, s=0, sample_rate=22050):
        self.s = s
        super().__init__(sample_rate=sample_rate,
                         function=self._offset,
                         s=s)
        self.descriptor = descriptors[__all__[2]]

    def _offset(self, y, s):

        sample_offs = int(self.sample_rate * abs(s))

        if len(y) <= sample_offs:
            raise ValueError("Offset longer than duration of sound!")
        if s < 0:
            return y[sample_offs:]
        else:
            return np.insert(y, 0, np.zeros(sample_offs))


class fadeAudio(_augmentation):
    """
    Create a logarithmic fade-in or fade-out for a sound.

    Parameters
    ----------
    `s`:
        Float. Fade time in seconds. Default is `0`.
    `direction`:
        String. Direction from which the fade is applied. Default
        is `"in"`.    
    `sample_rate`:
        Int. Sample rate of audio file (used for resampling).
        Default is `22050`
    """

    def __init__(self, s=0, direction="in", sample_rate=22050):
        self.s = s
        self.direction = direction
        if direction not in ["in", "out"]:
            raise ValueError(f"parameter '{direction}' not recognized!")

        super().__init__(sample_rate=sample_rate,
                         function=self._fade,
                         s=s,
                         direction=direction)
        self.s = s
        self.direction = direction
        self.descriptor = descriptors[__all__[3]]

    def _fade(self, y, s, direction):

        fade_len = self.sample_rate * s
        x_new = y

        if direction == "out":
            end = len(y)
            start = end - fade_len
            fade_curve = np.logspace(0, -3, fade_len)
            x_new[start:end] *= fade_curve
        else:
            fade_curve = np.logspace(-3, 0, fade_len)
            x_new[0:fade_len] *= fade_curve

        return x_new


class mixAudio(_augmentation):
    """
    Mix two wavefiles at random or specified timestamps.

    Parameters
    ----------
    `ratio`:
        Float. Ratio by which the sounds are mixed. 
        `0 <= ratio <= 1`, 1 ignores the noise, 0 the main sound. 
        Default is `1`.
    `start_at`:
        Float. Second at which the mixed in sound should be used.
        If unspecified, a random time will be used. Default is `None`.     
    `sample_rate`:
        Int. Sample rate of audio file (used for resampling).
        Default is `22050`

    Notes
    -----
    Augmented sound is as long as the original sound of interest, 
    not the mixed-in sound    
    """

    def __init__(self, ratio=0.5, start_at=None, sample_rate=22050):
        self.ratio = ratio
        self.start_at = start_at
        super().__init__(sample_rate=sample_rate,
                         function=self._mix,
                         ratio=ratio,
                         start_at=start_at)
        self.descriptor = descriptors[__all__[4]]

    def _mix(self, y, f_mix, ratio, start_at):
        y_mix, _ = lr.load(f_mix, sr=self.sample_rate)
        y_len = len(y)
        y_mix_len = len(y_mix)
        

        # use value of center sample as seed
        if start_at == None:
            rd_value = int(1000*y[int(y_len/2)])
            rd.seed(rd_value)
        else:
            start = int(start_at * self.sample_rate)    

        if y_len < y_mix_len:
            if start_at == None:
                start = rd.randint(0, y_mix_len-y_len)
            return y * ratio + y_mix[start:start+y_len] * (1-ratio)
        else:
            if start_at == None:
                start = rd.randint(0, y_len-y_mix_len)
                y_mix_pad = np.append(np.zeros(start), y_mix)
                y_mix_pad = np.append(
                    y_mix_pad, np.zeros(y_len-start-y_mix_len))

            return y * ratio + y_mix_pad * (1-ratio)
